/******************************************************************
*
* Licensed Materials - Property of IBM
*
* (C) Copyright IBM Corp. 2017, 2019. All Rights Reserved.
*
* US Government Users Restricted Rights - Use, duplication or
* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*
*******************************************************************
*/

/// \file ddl.hpp
/// \brief IBM Distributed Deep Learning Library

#ifndef __DDL_HPP__
#define __DDL_HPP__

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

#define DDL_MAJOR 1
#define DDL_MINOR 4
#define DDL_PATCH 0

/// List of reduction operations that DDL Supports
enum DDL_Op {
  DDL_OP_SUM = 0, ///< Summation
  DDL_OP_AVERAGE  ///< Average
};

#ifndef __cplusplus
typedef enum DDL_Op DDL_Op;
#endif

/// List of datatypes that DDL Supports
enum DDL_Type {
  DDL_TYPE_FLOAT = 0, ///< 32-bit Floating Point Number
  DDL_TYPE_HALF,      ///< 16-bit Floating Point Number
  DDL_TYPE_BYTE       ///< 8-bit  Value
};

#ifndef __cplusplus
typedef enum DDL_Type DDL_Type;
#endif

/// List of DDL statuses
enum DDL_Status {
  DDL_SUCCESS = 0, ///< Success
  DDL_FAILURE      ///< Unknown Failure
};

#ifndef __cplusplus
typedef enum DDL_Status DDL_Status;
#endif

/// List of DDL keys
enum DDL_Key {
  DDL_KEY_E2E_TIME = 0, ///< End2End runtime of DDL in sec
  DDL_KEY_JITTER,       ///< Jitter observed by DDL in sec
  DDL_KEY_DIM0_BW,      ///< Bandwidth@DIM observed by DDL in GB/s on dimension 0
  DDL_KEY_DIM1_BW,      ///< Bandwidth@DIM observed by DDL in GB/s on dimension 1
  DDL_KEY_DIM2_BW,      ///< Bandwidth@DIM observed by DDL in GB/s on dimension 2
  DDL_KEY_DIM3_BW,      ///< Bandwidth@DIM observed by DDL in GB/s on dimension 3
  DDL_KEY_DEVICE        ///< Device that DDL is running on. -1 indicates the CPU
};

#ifndef __cplusplus
typedef enum DDL_Key DDL_Key;
#endif

/// DDL Event
/// Represents an asynchronous DDL event
#ifndef __cplusplus
typedef struct DDL_Event DDL_Event;
#else
struct DDL_Event;
#endif

/// Initializes DDL and MPI.
/// \param options contains a string of options for DDL
/// \return The result status
DDL_Status ddl_init(const char* options);

/// Cleans up DDL datastructures and calls MPI_Finalize
/// \return The result status
DDL_Status ddl_finalize();


/// Performs an allreduce calculation.
/// \param buffer points to an array of data
/// \param n is the number of elements in buffer
/// \param type is the datatype of buffer. Supported datatypes are: DDL_TYPE_FLOAT, DDL_TYPE_HALF.
/// \param op is the reduction operation to perform. Supported reduction operations are: DDL_OP_SUM, DDL_OP_AVERAGE.
/// \return The result status
#ifdef __cplusplus
DDL_Status ddl_allreduce(void* buffer, size_t n, DDL_Type type = DDL_TYPE_FLOAT, DDL_Op op = DDL_OP_SUM);
#else
DDL_Status ddl_allreduce(void* buffer, size_t n, DDL_Type type, DDL_Op op);
#endif

/// Performs an asyncronous allreduce calculation.
/// Note: Only a single allreduce call can be in-flight at a time
/// \param buffer points to an array of data
/// \param n is the number of elements in buffer
/// \param type is the datatype of buffer. Supported datatypes are: DDL_TYPE_FLOAT, DDL_TYPE_HALF.
/// \param op is the reduction operation to perform. Supported reduction operations are: DDL_OP_SUM, DDL_OP_AVERAGE.
/// \param event will store the new DDL event
/// \return The result status
DDL_Status ddl_allreduce_async(void* buffer, size_t n, DDL_Type type, DDL_Op op, DDL_Event** event);

/// Waits until the specified asynchronous event has completed
/// \param event is the event to wait on
/// \return The result status
DDL_Status ddl_wait(DDL_Event* event);


// Memory Operations

/// Allocates memory on the GPU that can be used by DDL
/// \param buffer is a pointer to the memory that will be returned
/// \param align is the alignment to be used
/// \param n is the number of bytes being requested
/// \return The number of bytes that were allocated
size_t ddl_malloc(void** buffer, size_t align, size_t n);

/// Frees memory on the GPU
/// \param buffer is a pointer to the memory that will be freed
/// \return The result status
DDL_Status ddl_free(void* buffer);

/// Copies memory from the CPU to a GPU
/// \param gpu_dst is a pointer to the destination on the GPU
/// \param cpu_src is a pointer to the source on the CPU
/// \param size is the number of bytes to copy
/// \return The result status
DDL_Status ddl_memcpy_host_to_device(void* gpu_dst, const void* cpu_src, size_t size);

/// Copies memory from a GPU to the CPU
/// \param cpu_dst is a pointer to the destination on the CPU
/// \param gpu_src is a pointer to the source on the GPU
/// \param size is the number of bytes to copy
/// \return The result status
DDL_Status ddl_memcpy_device_to_host(void* cpu_dst, const void* gpu_src, size_t size);

/// Copy all data from an array of arrays into a single array.
/// \param destination is the array that will hold the concatenated data
/// \param sources is the array of arrays to concatenate
/// \param sizes is an array, the length of sources, that contains the number of bytes
///        in each array in sources
/// \param n is the number of arrays in sources
/// \return The result status
DDL_Status ddl_memcpy_gather(char* destination, char** sources, size_t* sizes, size_t n);

/// Copy a single array into an array of arrays
/// \param destinations is the array of arrays to copy to
/// \param sizes is an array, the length of destinations, that contains the number of bytes
///   in each array in destinations
/// \param n is the number of arrays in destinations
/// \param source is the array that will be copied into the destinations
/// \return The result status
DDL_Status ddl_memcpy_scatter(char** destinations, size_t* sizes, size_t n, char* source);

/// Performs a sync for every Cuda stream that DDL is using
/// \return The result status
DDL_Status ddl_memsync();

/// Performs a broadcast operation to all nodes in the cluster
/// \param buffer is the data to be broadcasted
/// \param n is the number of elements to be broadcasted
/// \param type is the datatype to be broadcasted. Supported datatypes are: DDL_TYPE_FLOAT, DDL_TYPE_HALF, DDL_TYPE_BYTE.
/// \param root is the rank of the node to broadcast from
/// \return The result status
DDL_Status ddl_bcast(void* buffer, int n, DDL_Type type, int root);

// Utilities

/// Obtains the rank of the current process
/// \param rank will contain the rank of the current node after execution
/// \return The result status
DDL_Status ddl_rank(int* rank);

/// Obtains the number of distributed processes
/// \param size will contain the number of distributed processes after execution
/// \return The result status
DDL_Status ddl_size(int* size);

/// Obtains the local rank of the current process
/// \param rank will contain the local rank of the current node after execution
/// \return the result status
DDL_Status ddl_local_rank(int* rank);

/// Obtains the number of local processes
/// \param size will contain the number of local processes after execution
/// \return the result status
DDL_Status ddl_local_size(int* size);

/// Obtains the value for the given key
/// \param key is the requested key
/// \param value will contain the value corresponding to the key
/// \return the result status
DDL_Status ddl_get_key_value(DDL_Key key, void* value);

/// Generate group ID for collective operations when running DDL without MPI
/// Note: ID needs to be passed to ddl_int function as the '-ddl_group_id' argument
///       and caller process needs to be alive when ddl_init is called
/// \param generated_id pre-allocated string of at least 173 bytes. 172-char ID + null char will be written.
/// \return the result status
DDL_Status ddl_generate_group_id(char* generated_id);

#ifdef __cplusplus
}
#endif

#endif
