import pyddl
import os
import subprocess
import sys
import numpy as np
import re

def init_and_run_allreduce_test(init):
    n = 100
    pyddl.init(init)

    initial_values = [x for x in range(n)]
    arr = np.array(initial_values, dtype=np.float32)
    res, buf = pyddl.malloc(64, n * 4)
    pyddl.memcpy_host_to_device(buf, arr)
    pyddl.memsync()

    pyddl.allreduce(buf, n, pyddl.DDL_TYPE_FLOAT, pyddl.DDL_OP_SUM)

    pyddl.memcpy_device_to_host(arr, buf)
    pyddl.memsync()
    pyddl.free(buf)

    # Verify that the allreduce produced the correct result
    result = np.array_equal(pyddl.size() * np.array(initial_values, dtype=np.float32), arr)

    if not result:
        sys.exit(1)

    # Generate ID for next call
    id_size = 172
    id_np = np.empty(id_size, dtype='ubyte')
    if pyddl.rank() == 0:
        next_id = pyddl.generate_group_id()
        id_np = np.array([ord(i) for i in next_id], dtype='ubyte')

    res, buf = pyddl.malloc(1, id_size)
    pyddl.memcpy_host_to_device(buf, id_np)
    pyddl.memsync()

    pyddl.bcast(buf, id_size, pyddl.DDL_TYPE_BYTE, 0)

    pyddl.memcpy_device_to_host(id_np, buf)
    pyddl.memsync()
    pyddl.free(buf)

    pyddl.finalize()

    return id_np.tobytes().decode("utf-8")

def update_init(init, worker_count = None, worker_index = None, device_index = None, id = None):
    if worker_count is not None:
        init = re.sub(r'-ddl_worker_count +([0-9]+)', '-ddl_worker_count %d' % worker_count, init)
    if worker_index is not None:
        init = re.sub(r'-ddl_worker_index +([0-9]+)', '-ddl_worker_index %d' % worker_index, init)
    if device_index is not None:
        init = re.sub(r'-ddl_device_index +([0-9]+)', '-ddl_device_index %d' % device_index, init)
    if id is not None:
        init = re.sub(r'-ddl_group_id +[^ ]+', '-ddl_group_id %s' % id, init)
    return init

def add_worker():
    init = os.environ['DDL_OPTIONS']
    print('Initializing and running allreduce with one more worker. Options: ' + init)
    init_and_run_allreduce_test(init)

def main():
    if len(sys.argv) > 1 and sys.argv[1] == '--new_worker':
        add_worker()
        sys.exit(0)
    init = os.environ['DDL_OPTIONS']
    print('Initializing and running allreduce with options: ' + init)
    next_id = init_and_run_allreduce_test(init)

    worker_count = int(re.search(r'-ddl_worker_count +([0-9]+)', init).group(1))
    worker_index = int(re.search(r'-ddl_worker_index +([0-9]+)', init).group(1))
    device_index = int(re.search(r'-ddl_device_index +([0-9]+)', init).group(1))

    # Remove the last worker for the next allreduce call
    if worker_index == worker_count - 1:
        sys.exit(0)
    worker_count = worker_count - 1
    init = update_init(init, worker_count=worker_count, id=next_id)
    print('Initializing and running allreduce with one less worker. Options: ' + init)
    next_id = init_and_run_allreduce_test(init)

    # Add another worker for the next allreduce call
    if worker_index == worker_count - 1:
        new_worker_init = update_init(init, worker_count=worker_count + 1,
                                      worker_index=worker_index + 1, device_index= device_index + 1,
                                      id=next_id)
        os.environ['DDL_OPTIONS'] = new_worker_init
        subprocess.Popen('python %s %s' % (__file__, '--new_worker'), shell=True,
                         universal_newlines=True)
    worker_count = worker_count + 1
    init = update_init(init, worker_count=worker_count, id=next_id)
    print('Initializing and running allreduce with one more worker. Options: ' + init)
    init_and_run_allreduce_test(init)

if __name__ == '__main__':
    main()
