"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

from __future__ import print_function

import os
import sys
import re
import argparse
import socket

import ddlrun.error as error
import ddlrun.mpi_checks as mpi_checks
import ddlrun.utilities as utilities
import ddlrun.rankfile as rankfile

ENV_VARS = {
    "PATH", "LD_LIBRARY_PATH", "PYTHONPATH", "DDL_GROUP_SIZE",
    "DDL_SYNC_METHOD", "CAFFE_CONDA_LIBPATH"
}


def make_parser():
    """Make the ArgParser object."""
    parser = argparse.ArgumentParser(
        description=
        'Provide a streamlined interface to launch programs which use the '
        'IBM Distributed Deep Learning library.')

    parser.add_argument(
        '-m',
        '--m',
        '-mode',
        '--mode',
        dest='mode',
        type=str,
        default=None,
        help='Specify MODE as the ddl mode to use.')

    parser.add_argument(
        '-ddlarg',
        '--ddlarg',
        '-ddloption',
        '--ddloption',
        dest='ddl_args',
        action='append',
        type=str,
        default=list(),
        help='Pass DDL_ARGS through as an argument to ddl.')

    parser.add_argument(
        '-no_ddloptions',
        '--no_ddloptions',
        dest='pass_ddl_options',
        action='store_false',
        help='Do not pass the ddloptions env var to mpi.')

    parser.add_argument(
        '-rf',
        '--rf',
        '-rankfile',
        '--rankfile',
        dest='rankfile',
        action='store',
        type=str,
        help=
        'Provide an MPI rank file. (Skips most error checking, including connection '
        'testing and key validation)')

    parser.add_argument(
        '-H',
        '--H',
        '-host',
        '--host',
        dest='host_list',
        action='store',
        type=str,
        help=
        '<host1,host2,...,hostN> List of hosts to use when generating a rankfile.'
    )

    parser.add_argument(
        '-hostfile',
        '--hostfile',
        dest='host_file',
        action='store',
        type=str,
        help='File containing list of hosts to use when generating a rankfile, '
        'with one host on each line.')

    parser.add_argument(
        '-aisles',
        '--aisles',
        dest='rank_args',
        metavar='AISLES',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help='Number of aisles. If no option given 1 is assumed.')

    parser.add_argument(
        '-racks',
        '--racks',
        dest='rank_args',
        metavar='RACKS',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help='Number of racks in each aisle. If no option given 1 is assumed.')

    parser.add_argument(
        '-nodes',
        '--nodes',
        dest='rank_args',
        metavar='NODES',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help=
        'Number of nodes per rack. If no option given the number of hosts given is assumed.'
    )

    parser.add_argument(
        '-accelerators',
        '--accelerators',
        dest='hw_args',
        metavar='ACCELERATORS',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help=
        'Number of accelerators per node. If no option given the number in the current '
        'machine is assumed. Error on values larger than 16.')

    parser.add_argument(
        '-sockets',
        '--sockets',
        dest='hw_args',
        metavar='SOCKETS',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help=
        'Number of CPU sockets per node. Values can be 1 or any even number 2 or larger. '
        'If no option given the number in the current machine is assumed.')

    parser.add_argument(
        '-cores',
        '--cores',
        dest='hw_args',
        metavar='CORES',
        action=RankgenArgAction,
        nargs=1,
        type=int,
        default=list(),
        help='Number of cores per socket. '
        'If no option given the number in the current machine is assumed.')

    parser.add_argument(
        '-all_cores',
        '--all_cores',
        dest='rank_args',
        action=RankgenArgAction,
        nargs=0,
        default=list(),
        help=
        'Each instance can use all cores. Default is to equally divide cores between instances.'
    )

    parser.add_argument(
        '-mpiarg',
        '--mpiarg',
        dest='mpi_args',
        action='append',
        type=str,
        default=list(),
        help='Pass MPI_ARGS through as an argument to the mpirun command.')

    parser.add_argument(
        '-tcp',
        '--tcp',
        '-TCP',
        '--TCP',
        dest='tcp',
        action='store_true',
        help=
        'Have mpi use TCP instead of infiniband to communicate between nodes.')

    parser.add_argument(
        '-port',
        '--port',
        dest='default_port',
        type=str,
        default='2200',
        help=
        'Lower bound of port range used for specific inter process communication. '
        'Upper bound is lower_bound + number_gpus_per_host - 1.')

    parser.add_argument(
        '-v',
        '--v',
        '-verbose',
        '--verbose',
        dest='verbose',
        action='store_true',
        help='Provide more verbose output.')

    parser.add_argument(
        '-s',
        '--s',
        '-skipchecks',
        '--skipchecks',
        dest='skipchecks',
        action='store_true',
        help='Skip all diagnostic checks.')

    parser.add_argument(
        '-t',
        '--t',
        '-topology',
        '--topology',
        '-topography',
        '--topography',
        dest='topology',
        type=str,
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-no_local_hostname',
        '--no_local_hostname',
        dest='use_local_hostname',
        action='store_false',
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-ignore_conda_env',
        '--ignore_conda_env',
        '-ice',
        '--ice',
        dest='ignore_conda_env',
        action='store_true',
        help=argparse.SUPPRESS
        #'Skip checking the conda environment and do not activate any conda environment.'
    )

    parser.add_argument(
        '-debug_timeout',
        '--debug_timeout',
        dest='debug_timeout',
        type=int,
        default=180,
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-n',
        '--n',
        '-np',
        '--np',
        dest='num_procs',
        action=InvalidParamAction,
        default=1,
        const=error.ErrorMessage.NUM_PROC_UNSUPPORTED,
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-rgarg',
        '--rgarg',
        '-rankgenarg',
        '--rankgenarg',
        action=InvalidParamAction,
        const=error.ErrorMessage.RANKGEN_PARAM_UNSUPPORTED,
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-flex',
        '--flex',
        dest='use_flex_mode',
        action='store_true',
        help='Run DDL without MPI.')

    parser.add_argument("program", type=list)

    return parser


def preprocess_args(parser):
    """Preprocess command line arguments to handle problems.
        -Add whitespaces to the arguments of ddl options, allowing single and double
        dash options.
        -At the first occurrence of a non-ddl argument, convert the remaining
        arguments into a list for the program arguments. This prevents argparse from
        trying to handle the args for the sub-program."""
    ddl_options = [
        key.strip('-') for key, arg in parser._option_string_actions.items()  #pylint: disable=protected-access
        if arg.nargs != 0
    ]
    ddl_flags = [
        key.strip('-') for key, arg in parser._option_string_actions.items()  #pylint: disable=protected-access
        if arg.nargs == 0
    ]
    copyarg = iter(enumerate(sys.argv))
    for i, arg in copyarg:
        if i == 0:
            continue
        elif '=' in arg:
            continue
        elif arg.lstrip('-') in ddl_options:
            sys.argv[i + 1] = " " + sys.argv[i + 1]
            next(copyarg, None)
        elif arg.lstrip('-') not in ddl_flags:
            cutoff = i
            new = sys.argv[cutoff:]
            sys.argv = sys.argv[:cutoff]
            sys.argv.append(new)
            break


def postprocess_args(args):
    """Undo some of the pre-processing where necessary.
        -Remove the leading whitespace from all string args.
        -Remove the leading whitespace from each mpiarg."""
    try:
        basestring
    except NameError:
        basestring = str  #pylint: disable=redefined-builtin
    for i in range(len(args.mpi_args)):
        args.mpi_args[i] = args.mpi_args[i].strip()
    for arg in vars(args):
        value = getattr(args, arg)
        if isinstance(value, basestring):
            setattr(args, arg, value.strip())
    return args


def make_initial_args():
    """Setup initial args for logging."""
    initial_args = argparse.Namespace()

    utilities.run_command("mkdir -m 777 -p /tmp/DDLRUN")
    _, initial_args.working_dir, _ = \
        utilities.run_command_capture(\
            "mktemp -d \"${TMPDIR:-/tmp/DDLRUN/}$(basename DDLRUN).XXXXXXXXXXXX\"")
    initial_args.working_dir = initial_args.working_dir.strip()
    initial_args.logfile = os.path.join(initial_args.working_dir, "ddlrun.log")

    error.make_logger(initial_args)

    return initial_args


def read_args(initial_args):  #pylint: disable=too-many-branches,too-many-statements
    """Parse program command line arguments.

    Args:
        args: Default arguments namespace object.

    Returns:
        Parsed program inputs.

    Raises:
        GenericDDLError
    """
    parser = make_parser()
    preprocess_args(parser)
    args = parser.parse_args()
    args.working_dir = initial_args.working_dir
    args.logfile = initial_args.logfile
    args.env_vars = []
    args = postprocess_args(args)

    args.lsf = "LSB_MCPU_HOSTS" in os.environ

    error.set_logging_level(args)

    validate_input(args)

    args.conda_env = None
    args.conda_setup_script = None

    if args.host_file is not None:
        args.host_list = rankfile.parse_hostfile(args)
    if args.lsf:
        if args.rankfile is None and args.host_file is None and args.host_list is None:
            args.host_list = rankfile.parse_lsf_hosts()
        else:
            error.log_warning(error.ErrorMessage.LSF_AND_HOSTS)

    args.hosts = args.host_list.split(
        ",") if not args.host_list is None else list()

    # Replace all local hostname variations with the advertised hostname
    local_hostname = socket.gethostname()
    args.hosts = [local_hostname if utilities.NetUtils.is_local_host(h) else h for h in args.hosts]

    for var in ENV_VARS:
        if var in os.environ:
            args.mpi_args.extend(["-x", var])
            args.env_vars.append(var + "=$" + var)
    if args.verbose:
        args.mpi_args.append("--report-bindings")

    args.use_spectrum = mpi_checks.check_spectrum()
    if args.use_spectrum and not args.tcp:
        args.mpi_args.append("-disable_gdr")
        args.mpi_args.append("-gpu")
    elif args.use_spectrum and args.tcp:
        args.mpi_args.append("-tcp")
        args.mpi_args.append("-disable_gpu_hooks")

    if args.tcp:
        args.mpi_args.extend(["-x", "NCCL_IB_DISABLE=1"])
        args.env_vars.append("NCCL_IB_DISABLE=1")

    if args.use_local_hostname and not args.hosts and args.rankfile is None:
        args.hosts.append(local_hostname.strip())

    args.mpirun_exe = mpi_checks.mpirun_test()

    return args


def validate_input(args):
    """Performs various input validation.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        GenericDDLError
    """
    if args.rankfile is not None and args.rank_args:
        raise error.GenericDDLError(error.ErrorMessage.RANK_ARG_FILE)
    if args.rankfile is not None and args.host_list is not None:
        raise error.GenericDDLError(error.ErrorMessage.RANK_FILE_HOST)
    if not args.pass_ddl_options and args.topology is not None:
        raise error.GenericDDLError(error.ErrorMessage.TOPO_NO_DDLOPTS)
    if not args.pass_ddl_options and args.mode is not None:
        raise error.GenericDDLError(error.ErrorMessage.MODE_NO_DDLOPTS)
    if not args.pass_ddl_options and args.ddl_args:
        raise error.GenericDDLError(error.ErrorMessage.DDLARGS_NO_DDLOPTS)
    if args.host_file is not None and args.host_list is not None:
        raise error.GenericDDLError(error.ErrorMessage.HOSTFILE_AND_HOSTLIST)
    if args.rankfile is not None and args.host_file is not None:
        raise error.GenericDDLError(error.ErrorMessage.HOSTFILE_AND_RANKFILE)
    if re.match(r"^\d+$", args.default_port) is None:
        raise error.GenericDDLError(error.ErrorMessage.INVALID_PORT,
                                    args.default_port)


def replace_cores(args, new_cores):
    """Replace cores in the args namespace."""
    cores_replaced = False
    for i in range(len(args.rank_args)):
        if "-cores" in args.rank_args[i]:
            args.rank_args[i + 1] = str(new_cores)
            cores_replaced = True
    if not cores_replaced:
        args.rank_args.extend(["--cores", str(new_cores)])
    return args


def handle_ddl_options(args):
    """Handle adding ddl options to the mpi_args."""
    if args.pass_ddl_options:
        mode_string = ""
        if args.mode is None:
            args.mode = "p"
        if args.topology is None:
            mode_string = args.mode
        else:
            mode_string = "%s:%s" % (args.mode, args.topology)

        args.mpi_args.extend([
            "-x",
            "\'DDL_OPTIONS=-mode %s %s\'" %
            (mode_string, utilities.make_arg_list(args.ddl_args))
        ])

        if "DDL_OPTIONS" in os.environ:
            error.log_warning(error.ErrorMessage.DDL_OPTIONS_SET)
    return args


class RankgenArgAction(argparse.Action):  #pylint: disable=too-few-public-methods
    """Argparse Action for Rankgen Args.

    All args which use this action will have their option string and input added
    to the rank_args list.

    Specifying the dest as hw_args indicates that this argument is required for the
    hardware check so it will also be added to the hw_args list.

    hw_args represent the subset of rank_args that need to be passed to the hardware check.

    """

    def __init__(self, option_strings, **kwargs):
        super(RankgenArgAction, self).__init__(
            option_strings=option_strings, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            if values[0] <= 0:
                raise error.GenericDDLError(
                    error.ErrorMessage.INVALID_RANKGEN_ARG, str(values[0]),
                    option_string)
            new_args = ["--" + option_string.lstrip('-'), str(values[0])]
        else:
            new_args = ["--" + option_string.lstrip('-')]
        if getattr(namespace, 'rank_args') is None:
            setattr(namespace, 'rank_args', new_args)
        else:
            getattr(namespace, 'rank_args').extend(new_args)
        if self.dest == 'hw_args':
            if getattr(namespace, self.dest) is None:
                setattr(namespace, self.dest, new_args)
            else:
                getattr(namespace, self.dest).extend(new_args)


class InvalidParamAction(argparse.Action):  #pylint: disable=too-few-public-methods
    """Argparse Action for Deprecated Args."""

    def __init__(self, option_strings, **kwargs):
        super(InvalidParamAction, self).__init__(
            option_strings=option_strings, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        raise error.GenericDDLError(self.const, option_string)
